"""tasks to create python distribution packages (distutils/setuptools)

Example:
 - create a `MANIFEST.in` file with all tracked files in a git repo
 - upload an `sdist` package

::

    from doitpy.package import Package

    def task_package():
        pkg = Package()
        yield pkg.manifest_git()  # package:manifest
        yield pkg.sdist()         # package:sdist
        yield pkg.sdist_upload()  # pypi:sdist_upload

"""

import subprocess


class Package(object):
    """helper to create tasks to upload a python package to PyPi"""

    def __init__(self):
        self.revision_file = None


    def revision_git(self, file_name='revision.txt'):
        """create file with repo rev number"""
        cmd = "git rev-list --branches=master --max-count=1 HEAD > {}"
        self.revision_file = file_name
        return {
            'basename': 'package',
            'name': 'revision',
            'actions': [cmd.format(file_name)],
            'targets': [file_name],
            }

    def _create_manifest_git(self):
        cmd = "git ls-tree --name-only -r HEAD"
        file_list = subprocess.check_output(cmd, shell=True, )
        with open('MANIFEST.in', 'w') as manifest:
            for filename in file_list.decode('utf-8').splitlines():
                manifest.write('include {}\n'.format(filename))


    def manifest_git(self):
        """create MANIFEST.in file for distutils/setuptools

        Put all files being tracked by git into the manifest
        """
        actions = [self._create_manifest_git]
        file_dep = []
        if self.revision_file:
            file_dep.append(self.revision_file)
            cmd = "echo 'include {}' >> MANIFEST.in"
            actions.append(cmd.format(self.revision_file))
        return {
            'basename': 'package',
            'name': 'manifest',
            'actions': actions,
            'file_dep': file_dep,
            'targets': ['MANIFEST.in'],
            'uptodate': [False],
            }

    def sdist(self):
        """create sdist package"""
        return {
            'basename': 'package',
            'name': 'sdist',
            'actions': ["python setup.py sdist"],
            'file_dep': ['MANIFEST.in'],
            'uptodate': [False],
            'verbosity': 2,
            }

    def sdist_upload(self):
        """upload sdist package to pypi"""
        return {
            'basename': 'pypi',
            'name': 'sdist_upload',
            'actions': ["python setup.py sdist upload"],
            'file_dep': ['MANIFEST.in', 'setup.py'],
            'verbosity': 2,
            }



'''
create tasks for coverage.py


Add coverage related tasks to ``dodo.py``::

    from doitpy.coverage import Coverage, PythonPackage

    def task_coverage():
        """show coverage for all modules including tests"""
        cov = Coverage([PythonPackage('my_pkg_name', 'tests')],
                       config={'branch':False, 'parallel':True,
                               'omit': ['tests/no_cover.py']},
                       )
        yield cov.all() # create task `coverage`
        yield cov.src() # create task `coverage_src`
        yield cov.by_module() # create tasks `coverage_module:<path/to/test>`


'''

import glob

from configclass import Config


def sep(*args):
    """join strings or list of strings ignoring None values"""
    return ' '.join(a for a in args if a)


class PythonFiles(object):
    """python code: a module or a package"""

    def all_modules(self):
        """Yield all source and test modules."""
        for mod in self.src + self.test:
            yield mod


class PythonModule(PythonFiles):
    """reference to a single python module / test for the module

    :ivar list-str src: list of path of source modules
    :ivar list-str test: list of path of all modules from test folder

    """
    def __init__(self, path, test_path):
        """
        :param str/pathlib.Path path: path to python module.
        :param str/pathlib.Path test_path: path to module with tests.
        """
        self.src = [path]
        self.test = [test_path]


class PythonPackage(PythonFiles):
    """Contain list of modules of the package (does not handle sub-packages)

    :ivar str src_dir: path to dir containing package modules
    :ivar str test_dir: path to dir containing package tests
    :ivar list-str src: list of path of source modules
    :ivar list-str test: list of path of all modules from test folder
    """

    # TODO should track sub-packages

    #: :class:`confclass.Config`
    #:
    #: :var str test_prefix: string prefix on name of files
    #: :var str pkg_test_dir: path to location of test files
    config = Config(
        test_prefix = 'test_',
        pkg_test_dir = 'tests',
        )

    def __init__(self, path, test_path=None, config=None):
        """
        :param str/pathlib.Path path: dir path to package.
        :param str/pathlib.Path test_path: if test_path is not given assume
                it is on config.pkg_test_dir inside source package.
        """
        self.config = self.config.make(config)
        self.test_prefix = self.config['test_prefix']

        self.src_dir = str(path) if path else ''
        if test_path is None:
            self.test_dir = '{}/{}'.format(self.src_dir,
                                            self.config['pkg_test_dir'])
        else:
            self.test_dir = str(test_path)
        self.src = glob.glob("{}/*.py".format(self.src_dir))
        self.test = glob.glob("{}/*.py".format(self.test_dir))


class Coverage(object):
    """generate tasks for coverage.py"""

    #: :class:`confclass.Config`
    #:
    #: :var string cmd_run_test: shell command used to run tests
    #: :var branch bool: measure branche coverage
    #: :var parallel bool: measure using `--parallel-mode` (needed for
    #:              subprocess and multiprocess coverage
    #: :var concurrency str: --concurrency library
    #: :var list-str omit: list of paths to omit from coverage
    config = Config(
        cmd_run_test = "`which py.test`",
        branch=True,
        parallel=False,
        concurrency='',
        omit=[])

    def __init__(self, pkgs, config=None):
        """
        :param list-PythonFiles pkgs: packages/modules to measure coverage
        """
        self.config = self.config.make(config)
        self.pkgs = []
        for pkg in pkgs:
            assert isinstance(pkg, PythonFiles)
            self.pkgs.append(pkg)


    def _action_list(self, modules, test=None):
        """return list of actions to be used in a doit task"""
        run_options = []
        if self.config['branch']:
            run_options.append('--branch')
        if self.config['parallel']:
            run_options.append('--parallel-mode')
        if self.config['concurrency']:
            run_options.append('--concurrency')
            run_options.append(self.config['concurrency'])

        report_options = []
        if self.config['omit']:
            omit_list = ','.join(self.config['omit'])
            report_options.append('--omit {}'.format(omit_list))

        actions = [sep("coverage run", sep(*run_options),
                       self.config['cmd_run_test'], test)]
        if self.config['parallel']:
            actions.append('coverage combine')
        actions.append(sep("coverage report --show-missing",
                           sep(*report_options), sep(*modules)))
        return actions


    def all(self, basename='coverage'):
        """show coverage for all modules including tests"""
        all_modules = []

        for pkg in self.pkgs:
            for module in pkg.all_modules():
                all_modules.append(module)

        yield {
            'basename': basename,
            'actions': self._action_list(all_modules),
            'verbosity': 2,
            }


    def src(self, basename='coverage_src'):
        """show coverage for all modules (exclude tests)"""
        all_modules = []

        for pkg in self.pkgs:
            for module in pkg.src:
                all_modules.append(module)

        yield {
            'basename': basename,
            'actions': self._action_list(all_modules),
            'verbosity': 2,
            }


    def by_module(self, basename='coverage_module'):
        """show coverage for individual modules"""
        for pkg in self.pkgs:
            prefix = '{}/{}'.format(pkg.test_dir, pkg.test_prefix)
            to_strip = len(prefix)
            tests = glob.glob('{}*.py'.format(prefix))
            for test in tests:
                source = pkg.src_dir + '/' + test[to_strip:]
                yield {
                    'basename': basename,
                    'name': test,
                    'actions': self._action_list([source, test], test),
                    'verbosity': 2,
                    }




"""create tasks related to documentation generation

Example to create tasks to:
 - spell check all restructured text files (including README)
 - build sphinx docs in `doc` folder
 - upload built sphinx docs to `pythonhosted` using `setuptools`

::

    def task_docs():
        doc_files = glob.glob('doc/*.rst') + ['README.rst']
        yield docs.spell(doc_files, 'doc/dictionary.txt')
        yield docs.sphinx('doc/', 'doc/_build/html/', task_dep=['spell'])
        yield docs.pythonhosted_upload('doc/_build/html/', task_dep=['sphinx'])


"""

import subprocess

def check_no_output(doc_file, dictionary):
    """run spell checker on file

    spell always return successful code (0)
    so this checks if the output is empty
    """
    # -l list misspelled words
    # -p set path of personal dictionary
    cmd = 'hunspell -l -d en_US -p {} {}'.format(dictionary, doc_file)
    output = subprocess.check_output(cmd, shell=True,
                                     universal_newlines=True)

    if len(output) != 0:
        print(output)
        return False
    else:
        return True

# task creator
def spell(files, dictionary):
    """spell checker for doc files

    :param list-str files: list of files to spell check
    :param str dictionary: path of dictionary with extra words
    """
    for doc_file in files:
        yield {
            'basename': 'spell',
            'name': doc_file,
            'actions': [(check_no_output, (doc_file, dictionary))],
            'file_dep': [dictionary, doc_file],
            'verbosity': 2,
            }


# task creator
def sphinx(root_path, build_path, sphinx_opts='', task_dep=None):
    """build sphinx docs

    :param str root_path: root path of sphinx docs
    :param str build_path: path generated sphinx docs will be saved in
    :param str sphinx_opts: `sphinx-build` command line options
    :param list-str task_dep: list of tasks this task will depend on
    """
    cmd = "sphinx-build -b html {opts} -d {root}_build/doctrees {root} {build}"
    action = cmd.format(root=root_path, build=build_path, opts=sphinx_opts)
    # sphinx has its own check it up-to-date so we dont care
    # about always re-executing the task.
    task = {
        'basename': 'sphinx',
        'actions': [action],
        'verbosity': 2,
        }
    if task_dep:
        task['task_dep'] = task_dep
    yield task


# task creator
def pythonhosted_upload(www_path, task_dep):
    """deploy website (sphinx docs)

    :param str www_path: path to folder containig www files
    :param list-str task_dep: list of tasks this task will depend on
    """
    action = "python setup.py upload_docs --upload-dir %s"
    yield {
        'basename': 'pythonhosted_upload',
        'actions': [action % www_path],
        'task_dep': task_dep,
        'verbosity': 2,
        }



"""
Helper to create tasks that execute pyflakes.


Example to create a task for every python module (found recursively from current
path), excluding test modules and also `doc/conf.py`.

::

    from doitpy.pyflakes import Pyflakes

    def task_pyflakes():
        flaker = Pyflakes(exclude_patterns=['test_*'])
        yield flaker.tasks('**/*.py', exclude_paths=['doc/conf.py'])


"""

from __future__ import absolute_import

from pathlib import Path
from pyflakes.api import checkPath
from pyflakes.checker import Checker

from configclass import Config


def check_path(filename):
    """a doit action - execute pyflakes in a single file.
    :return bool: check succeded
    """
    return not bool(checkPath(filename))


class Pyflakes(object):
    """generate tasks for pyflakes
    """

    #: :class:`confclass.Config`
    #:
    #: :var str base_dir: list of path patterns of files to be linted
    #: :var list-str exclude_patterns: list of pattern of files to be removed
    #:                                from selection
    #: :var list-str exclude_paths: list of path of files to be removed
    #:                             from selection
    config = Config(
        base_dir='.',
        exclude_patterns=[],
        exclude_paths=[],
        )


    def __init__(self, **kwargs):
        """:param kwargs: config params
        """
        self.config = self.config.make(kwargs)


    def __call__(self, py_file):
        """Return a task for single file.

        :param str pyfile: path to file
        :return: task metadata to run pyflakes on a single module
        """
        # 'unicode' is a builtin in py2 but not on py3.
        # Make sure pyflakes consider 'unicode' as a builtin so
        # it does not fail on py3.
        Checker.builtIns.add('unicode')
        return {
            'name': py_file,
            'actions': [(check_path, [py_file])],
            'file_dep': [py_file],
            }

    def tasks(self, pattern, **kwargs):
        """run pyflakes on python module

        yield one task for each file as given by pattern

        :param str pattern: path pattern of files to be linted
        """

        config = self.config.make(**kwargs)
        # yield a task for every py file in selection
        base = Path(config['base_dir'])
        excluded = set([base.joinpath(e) for e in config['exclude_paths']])
        for src in base.glob(pattern):
            if src in excluded:
                continue
            for exclude_pattern in config['exclude_patterns']:
                if src.match(exclude_pattern):
                    break
            else:
                yield self(str(src))
